"""
This module contains helpers that are used internally by Runium.
"""
from runium.constants import TIME_SCALES


def get_seconds(interval):
    """
    Returns seconds which is time * scale.
    If interval is a number or None then it will be returned as is.

    interval: Can be a number or a string in the following format: 'time scale'
    ie: '10 minutes'.
    """
    if type(interval) is str:
        interval_array = interval.split(' ')
        try:
            seconds = int(interval_array[0]) * TIME_SCALES[interval_array[1]]
        except KeyError:
            raise ValueError(
                "Valid time scales are: {}".format(list(TIME_SCALES.keys()))
            )
        except ValueError:
            raise ValueError(
                '''
                Invalid format in time scale. It must include an integer
                followed by a string.
                Try a format like this: '1 minute'.
                '''
            )
        else:
            return seconds

    elif isinstance(interval, (int, float)) or interval is None:
        return interval

    else:
        raise TypeError(
            "Time intervals can only be a string, an integer or a float."
        )



import pytest
import time
from runium.core import Runium


@pytest.fixture(scope="module")
def rnt():
    return Runium()


@pytest.fixture(scope="module")
def rnp():
    return Runium(mode='multiprocessing')


def test_run_3_times(rnt, rnp):
    rt = rnt.run(runnium_param, times=3).result()
    rp = rnp.run(runnium_param, times=3).result()
    assert rt['iterations'] == 3
    assert rp['iterations'] == 3


def test_loop_drift(rnt, rnp):
    prev_time = time.time()
    rnt.run(simple_task, every=0.1, times=10).result()
    time_elapsed = time.time() - prev_time
    assert time_elapsed < 1
    assert time_elapsed > 0.88
    prev_time = time.time()
    rnp.run(simple_task, every=0.1, times=10).result()
    time_elapsed = time.time() - prev_time
    assert time_elapsed < 1
    assert time_elapsed > 0.88


def test_start_in(rnt, rnp):
    prev_time = time.time()
    rnt.run(simple_task, start_in=0.1).result()
    time_elapsed = time.time() - prev_time
    assert time_elapsed < 0.11
    assert time_elapsed > 0.09
    prev_time = time.time()
    rnp.run(simple_task, start_in=0.1).result()
    time_elapsed = time.time() - prev_time
    assert time_elapsed < 0.11
    assert time_elapsed > 0.09


def test_skipping_tasks(rnt, rnp):
    prev_time = time.time()
    rnt.run(sleepy_task, every=0.1, times=2).result()
    time_elapsed = time.time() - prev_time
    assert time_elapsed < 0.51
    assert time_elapsed > 0.2
    prev_time = time.time()
    rnp.run(sleepy_task, every=0.1, times=2).result()
    time_elapsed = time.time() - prev_time
    assert time_elapsed < 0.51
    assert time_elapsed > 0.2


def test_kwargs(rnt, rnp):
    rt = rnt.run(
        task_with_kwargs, kwargs={'msg': 'Spam, Spam, Spam, egg and Spam'}
    ).result()
    rp = rnp.run(
        task_with_kwargs, kwargs={'msg': 'Spam, Spam, Spam, egg and Spam'}

    ).result()
    assert rt == 'Spam, Spam, Spam, egg and Spam'
    assert rp == 'Spam, Spam, Spam, egg and Spam'


def test_return_exception(rnt, rnp):
    rt = rnt.run(task_with_exception)
    rp = rnp.run(task_with_exception)
    pytest.raises(ValueError, rt.result)
    pytest.raises(ValueError, rp.result)


def test_detect_exception(rnt, rnp):
    assert str(rnt.run(task_with_exception).exception()) == 'ni'
    assert str(rnp.run(task_with_exception).exception()) == 'ni'


def test_run_no_kwargs_bleeding(rnt, rnp):
    assert rnt.run(simple_task).result() is True
    assert rnp.run(simple_task).result() is True


# DUMMY TASKS =================================================================
def simple_task():
    return True


def task_with_kwargs(msg=None, **kwargs):
    return msg


def runnium_param(runium):
    return runium


def task_with_exception(runium):
    raise ValueError('ni')
    return runium


def sleepy_task():
    time.sleep(0.2)
    return True



"""
This is the main module.
"""

import time
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from inspect import signature
import atexit
import uuid
from runium.util import get_seconds


class Runium(object):
    """
    This is the main class that gets instantiated by the user.
    """

    def __init__(self, mode='multithreading', workers=None):
        self.__tasks_list = {}
        self.workers = workers
        self.mode = mode
        if self.mode == 'multiprocessing':
            self._executor = ProcessPoolExecutor(self.workers)
        elif self.mode == "multithreading":
            self._executor = ThreadPoolExecutor(self.workers)
        else:
            raise ValueError(
                'mode can only be multiprocessing or multithreading.'
            )

        atexit.register(self._executor.shutdown)

    def run(
        self, task, every=None, times=None, start_in=0, kwargs={},
        callbacks=None
    ):
        """
        Creates a new Task, adds it to the tasks list and submits it to the
        PoolExecutor.
        Returns the Task object.
        """
        every = get_seconds(every)
        start_in = get_seconds(start_in)
        every, times = self.__set_every_times_defaults(every, times)

        task_id = uuid.uuid4().int
        future = self._executor.submit(
            _run_task,
            task, task_id, every, times, start_in, kwargs
        )

        current_task = _Task(task_id, future, callbacks)
        self.__add_task_to_tasks_list(current_task)

        return current_task

    def __add_task_to_tasks_list(self, task):
        self.__tasks_list[task.id] = task

    def __remove_task_from_tasks_list(self, task_id):
        try:
            del self.__tasks_list[task_id]
        except KeyError:
            pass

    def tasks(self):
        """
        Cleans up and returns the tasks list.
        """
        tasks_to_remove = []
        for task_id, task in self.__tasks_list.items():
            if task.future.done() is True:
                tasks_to_remove.append(task_id)
        for t_id in tasks_to_remove:
            self.__remove_task_from_tasks_list(t_id)

        return self.__tasks_list

    def __set_every_times_defaults(self, every, times):
        """
        Sets the :every and :times properties to sensible defaults if one or
        any of them are not set.
        Sets defaults so that Runium will:
            Run the task one time if the :times and :every interval are not
            set.
            Loop indefinitely if :every is set and :times is not set.
        """
        if every is None and times is None:
            times = 1
            every = 0
        elif every is not None and times is None:
            times = 0
        elif every is None and times is not None:
            every = 0
        return every, times


def _run_task(fn, id, interval, times, start_in, kwargs):
    """
    Runs the task, optionally for :times every :every seconds in :start_in
    seconds.
    This is blocking and is meant to be called by Runium in a new Thread or
    Process.
    """
    task_result = None
    iterations = 0

    if start_in > 0:
        time.sleep(start_in)

    next_time = time.time() + interval
    while True:
        iterations += 1

        # Runium will pass some stats and functions in the callable's runium
        # attribute.
        if 'runium' in signature(fn).parameters.keys():
            runium_param = _make_runium_param(iterations, times)
            task_result = fn(runium=runium_param, **kwargs)
        else:
            task_result = fn(**kwargs)

        if times > 0 and iterations >= times:
            break

        # Skip tasks if we are behind schedule:
        if interval > 0:
            next_time +=\
                (time.time() - next_time) // interval * interval + interval
            time.sleep(max(0, next_time - time.time()))

    return task_result


def _make_runium_param(iterations, times):
    """
    Creates and returns a dict with runium specific stats and functions that
    can be accessed from within the task.
    """
    context = {
        'iterations': iterations,
        'iterations_remaining': times - iterations
    }
    return context


class _Task(object):
    """
    This object represents the task that will be executed by Runium.
    It encapsulates the future object.
    """
    def __init__(self, id, future, callbacks):
        self.id = id
        self.future = future
        self.callbacks = callbacks
        self.then_run(callbacks)

    def then_run(self, callbacks):
        """
        Attach one or a list of callbacks to the task. If the task is finished
        the callbacks wi be executed imidiately in the order they are
        submitted.
        """
        if callbacks is not None:
            if isinstance(callbacks, list):
                for c in callbacks:
                    self.future.add_done_callback(c)
            else:
                self.future.add_done_callback(callbacks)
        return self

    def result(self, timeout=None):
        return self.future.result(timeout)

    def exception(self, timeout=None):
        return self.future.exception(timeout)

    def get_state(self):
        return self.future._state



import pytest
from runium.util import get_seconds


class TestGetSeconds(object):

    def test_number(self):
        assert get_seconds(10) == 10

    def test_float(self):
        assert get_seconds(1.2) == 1.2

    def test_seconds_string(self):
        assert get_seconds('10 seconds') == 10

    def test_minutes_string(self):
        assert get_seconds('10 minutes') == 600

    def test_hours_string(self):
        assert get_seconds('10 hours') == 36000

    def test_days_string(self):
        assert get_seconds('10 days') == 864000

    def test_invalid_string(self):
        pytest.raises(ValueError, get_seconds, 'abcd')

    def test_invalid_scale_string(self):
        pytest.raises(ValueError, get_seconds, '10 months')

    def test_invalid_type(self):
        pytest.raises(TypeError, get_seconds, [10])

    def test_none(self):
        assert get_seconds(None) is None



import os
import codecs
from setuptools import setup, find_packages


CURRENT_VERSION = "0.1.6"


def read_file(filename):
    """
    Read a utf8 encoded text file and return its contents.
    """
    with codecs.open(filename, 'r', 'utf8') as f:
        return f.read()


setup(
    name="runium",
    version=CURRENT_VERSION,
    author="Grigoris Chatzinikolaou",
    author_email="greghatzis@gmail.com",
    description="Clean and simple task concurrency for Python.",
    long_description=read_file('README.md'),
    keywords=[
        'concurrecy', 'asynchonous', 'tasks', 'scheduling', 'scheduler',
        'cron', 'multiprocessing', 'multithreading', 'callbacks',
        'non-blocking'
    ],
    url="https://github.com/AGTGreg/runium",
    license='MIT',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)
